! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

program ersensmap

!------------------------------------------------------------------------------
! eROSITA sensitivity maps.
! Author : GLamer, AIP   
! 
!
!
!------------------------------------------------------------------------------
  

  use param
  use ersensmap_mod
  use numkind
  use fitsimage
  use detect_io
  use cal
!  use evtool_mod
  use evtool_eventfile_mod
  use slib_constants



  implicit none


  type(ersensmap_param_type)                    :: par
  type(tEventFile), dimension(:), allocatable   :: eventlist
  type(pixlist_type), dimension(:), allocatable :: pixlist
  type(FPSF)                                    :: allStackPSF,allStackPSF2
  type(FPSF)                                    :: cluPSF, cluPSF2


  type(l1cube_type)                        :: detmask
  type(r4cube_type)                        :: expimage
  type(r4cube_type)                        :: bkgimage
  type(r4cube_type)                        :: srcimage
  type(r4image_type)                       :: sensmap
  type(area_table_type)                    :: areatab
  type(cal_par_type)                       :: cal_par
  
  integer(kind=int32)                      :: ierr,status
  integer(kind=int32)                      :: xsize,ysize

  character(len=*), parameter :: rtname='ERSENSMAP'

  status = 0
  ierr=0


!  write(*,*) 'main,ncol:',areatab%ncol

 
  ! open parameter and log files:
  call ersensmap_ini(par,status)
  ierr=max(ierr,status)  


  ! get task parameters:
  if (ierr < 3) then
    call ersensmap_par(par,status)
    ierr=max(ierr,status)
  endif


  ! get input images and source list:
  if (ierr < 3) then
    if (par%shapelet_flag) then
      call alloc_eventfile(eventlist, par%n_bkgimages, status)
      if (status > 0) then
        ierr=3
        call  wrerr(rtname,status,'Cannot allocate memory for event lists')
      endif
      allocate(pixlist(par%n_bkgimages))
    endif
    call ersensmap_in(par,expimage,bkgimage,detmask,status)
    ierr=max(ierr,status)
  endif


!  if  (par%shapelet_flag) then
!    if (ierr < 3) then
!      ! shape v1.8, allocated in 'slib_run'
!      !allocate(allStackPSF(nim))
!      !allocate(allStackPSF2(nim))
!      call ermldet_shapelets(par,boxlist,image,eventlist,allStackPSF,0,cluPSF,status)
!      call ermldet_shapelets(par,boxlist,image,eventlist,allStackPSF2,srcmodrad,cluPSF2,status)
!      do iim=1,par%n_bkgimages
!        call init_pixlist(eventlist(iim)%events,image%wcshdr(iim),pixlist(iim),status)
!      enddo  
!    endif
!    if (status > 0) then
!       ierr=3
!       call  wrerr(rtname,status,'Error preparing shapelet PSF')
!    endif   
!  endif 

  if (ierr < 3) then
    xsize=bkgimage%shapehdr(1)%naxis1
    ysize=bkgimage%shapehdr(1)%naxis2
    allocate(sensmap%arr(xsize,ysize),stat=status)
    if (status > 0) then
      ierr=3
      call  wrerr(rtname,status,'Cannot allocate memory for sensitivity map')
    endif   
  endif
 

  sensmap%shapehdr = bkgimage%shapehdr(1)
  sensmap%stdhdr   = bkgimage%stdhdr(1)
  sensmap%wcshdr   = bkgimage%wcshdr(1)




  if (ierr < 3) then
!    write(*,*) 'method=',par%method
    if (par%method == "FIT" .or. par%method == "fit") then 
      call ersensmap_psf(par,bkgimage,expimage,detmask,sensmap,status)
      ierr=max(ierr,status)
    else
      call ersensmap_calc(par,bkgimage,expimage,detmask,sensmap,status)
      ierr=max(ierr,status)
    endif
  endif
  
!  write(*,*) 'area_flag:',par%area_flag,ierr

  if (ierr < 3 .and. par%area_flag) then
    call area_calc(par,sensmap,detmask,areatab,status)
    ierr=max(ierr,status)
  endif
  
 
 
    !  call fput_mllist(par%mllistset,mllist,status)  
  if (ierr < 3) then
   call ersensmap_out(par,sensmap,areatab,status)
   ierr=max(ierr,status)
  endif

  
  status=0      
  ! close parameter and log files:   
  call  cpast(rtname,par%lun,ierr,status)


    if(status /= 0) then
      call wrerr(rtname,status,'Error closing parameter/log files')
      ierr=3
    endif    


end program ersensmap

