! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module ersensmap_mod

  use param
  use numkind  
  use fitsimage
  use detect_io
  use cal
  use psflib
  use fftpsf
  use slib_constants

!  use evtool_eventfile_mod
!  use evtool_events_mod
!  use slib_constants
!  use slib_engine
!  use slib_basic_calcs
!  use slib_data_evt
!  use slib_calc_space

  implicit none

  type area_row_type
    !sequence
    real(kind=real32)                      ::   limflux
    real(kind=real32)                      ::   sky_area
  end type area_row_type

  type area_table_type
    !sequence
    character(len=clong)                       :: filename
    character(len=16)                          :: extname = 'SKY_AREA'
    type(shapehdr_type)                        :: shapehdr
    type(stdhdr_type)                          :: stdhdr
    type(area_row_type), pointer, dimension(:) :: data  => NULL() 
    integer(kind=int32)                        :: ncol=2
    character(len=16), dimension(2) :: cname = [ character(len=16) :: &
                                                 'LIMFLUX','SKY_AREA']
    character(len=4) , dimension(2) :: cform=[ character(len=4)  ::  '1E','1E'] 
    character(len=16), dimension(33) :: cunit=[ character(len=16) :: 'erg/(cm^2 s)','deg^2']
  end type  area_table_type  

  

  type ersensmap_param_type

    !sequence
    integer(kind=int32)                     ::  lun
    character(len=clong),dimension(max_ima) ::  expimages
    integer(kind=int32)                     ::  n_expimages
    character(len=clong),dimension(max_ima) ::  detmasks
    integer(kind=int32)                     ::  n_detmasks
    character(len=clong),dimension(max_ima) ::  bkgimages
    integer(kind=int32)                     ::  n_bkgimages
    character(len=clong),dimension(max_ima) ::  srcimages
    integer(kind=int32)                     ::  n_srcimages
    character(len=clong)                    ::  sensimage
    character(len=clong)                    ::  area_table
    real(kind=real32),dimension(max_ima)    ::  emin
    integer(kind=int32)                     ::  n_emin
    real(kind=real32),dimension(max_ima)    ::  emax
    integer(kind=int32)                     ::  n_emax
    real(kind=real32),dimension(max_ima)    ::  ecf
    character(len=80)                       ::  method
    character(len=80)                       ::  aper_type
    real(kind=real32)                       ::  aper_size
    integer(kind=int32)                     ::  n_ecf
    real(kind=real32)                       ::  likemin
    real(kind=real32)                       ::  extlikemin
    real(kind=real32)                       ::  ext
    character(len=80)                       ::  extentmodel
    logical(kind=4)                         ::  detmask_flag
    logical(kind=4)                         ::  shapelet_flag
    logical(kind=4)                         ::  photon_flag
    logical(kind=4)                         ::  ext_flag
    logical(kind=4)                         ::  extlike_flag
    logical(kind=4)                         ::  area_flag
    logical(kind=4)                         ::  compress_flag

!    real(kind=real32)                       ::  cutrad  
!    real(kind=real32)                       ::  eef
!    logical(kind=4)                         ::  eef_flag   
  end type  ersensmap_param_type 



    
contains

!-------------------------------------------------------------------------------

  subroutine ersensmap_ini(par,ierr)

  !---------------------------------------------------------
  ! Calls subroutine OPAST to open parameter and log files
  !---------------------------------------------------------


    type(ersensmap_param_type), intent(out) :: par
    integer(kind=int32) , intent(out) :: ierr
    integer(kind=int32)               :: status
    character(len=*),parameter        :: rtname='ERSESNMAP_INI'
    character(len=*),parameter        :: taskname='ERSESNMAP' 

    ierr=0
    status=0


    call opast(taskname,par%lun,status)

    if (status /= 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening parameter file')
    endif 

  end subroutine ersensmap_ini


!-------------------------------------------------------------------------------

  subroutine ersensmap_par(par,ierr)
 
  !---------------------------------------------------------------------
  ! Get task parameters 
  !
  !---------------------------------------------------------------------    

    type(ersensmap_param_type) , intent(out)  :: par
    integer(kind=int32)      , intent(out)  :: ierr
    integer(kind=int32)                     :: status

    character(len=*),parameter         :: rtname='ERSENSMAP_PAR'
    
    

    ierr = 0


    call rgetca(par%lun,'EXPIMAGES',max_ima,par%expimages,par%n_expimages,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetca(par%lun,'DETMASKS',max_ima,par%detmasks,par%n_detmasks,status)
    if (status /= 0) ierr = max(ierr,status)           
    call rgetca(par%lun,'BKGIMAGES',max_ima,par%bkgimages,par%n_bkgimages,status)
    if (status /= 0) ierr = max(ierr,status)  
!    call rgetca(par%lun,'SRCIMAGES',max_ima,par%srcimages,par%n_srcimages,status)
!    if (status /= 0) ierr = max(ierr,status)  
    call rgetc(par%lun,'SENSIMAGE',par%sensimage,status)
    if (status /= 0) ierr = max(ierr,status)
    call rgetc(par%lun,'AREA_TABLE',par%area_table,status)
    if (status /= 0) ierr = max(ierr,status)
    call rgetfa(par%lun,'EMIN',max_ima,par%emin,par%n_emin,status)  
    if (status /= 0) ierr = max(ierr,status) 
    call rgetfa(par%lun,'EMAX', max_ima,par%emax,par%n_emax,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetfa(par%lun,'ECF', max_ima,par%ecf,par%n_ecf,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetc(par%lun,'METHOD',par%method,status)
    if (status /= 0) ierr = max(ierr,status)
    call rgetc(par%lun,'APER_TYPE',par%aper_type,status)
    if (status /= 0) ierr = max(ierr,status)
    call rgetf(par%lun,'APER_SIZE',par%aper_size,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetf(par%lun,'LIKEMIN',par%likemin,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetf(par%lun,'EXTLIKEMIN',par%extlikemin,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetf(par%lun,'EXT',par%ext,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetc(par%lun,'EXTENTMODEL',par%extentmodel,status)
    if (status /= 0) ierr = max(ierr,status)  
    call rgetl(par%lun,'DETMASK_FLAG',par%detmask_flag,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'SHAPELET_FLAG',par%shapelet_flag,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'PHOTON_FLAG',par%photon_flag,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'AREA_FLAG',par%area_flag,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'EXT_FLAG',par%ext_flag,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'EXTLIKE_FLAG',par%extlike_flag,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'COMPRESS_FLAG',par%compress_flag,status)
    if (status /= 0) ierr = max(ierr,status)   




 
     
    if (ierr /= 0)  call wrerr(rtname,ierr,'Error getting parameters')
                       
  end subroutine ersensmap_par

!-------------------------------------------------------------------------------

  subroutine ersensmap_in(par,expimage,bkgimage,detmask,ierr)

  !--------------------------------------------------------------------------------
  !
  ! Read fits images:
  ! If par%compress_flag == true, only relevant section of fits images 
  ! as specified in detection mask header, is applied:
  !   - use fget_<type>compress for reading
  !   - set global variables ximbuf,yimbuf, ixb0, iyb0
  ! 
  !--------------------------------------------------------------------------------


    type(ersensmap_param_type), intent(in)          :: par
!    type(i4cube_type)                               :: image
    type(l1cube_type)                               :: detmask
    type(r4cube_type)                               :: expimage
    type(r4cube_type)                               :: bkgimage
    integer(kind=int32), intent(out)                :: ierr
    integer(kind=int32)                             :: status,ext
    integer(kind=int32)                             :: i,j,k,ipoint,nrows,ord=1,nbands,nmask
    integer(kind=int32),dimension(:),allocatable    :: isort
    type(dethdr_type),dimension(max_exp)            :: mask_dethdr
    type(stdhdr_type),dimension(max_exp)            :: mask_stdhdr
    type(stdhdr_type),dimension(max_ima)            :: image_stdhdr
    logical(kind=4)                                 :: has_instsum,has_bandsum                              
    character(len=*), parameter                     :: rtname='ERSENSMAP_IN'
    character(len=80)                               :: errstr


    status=0
    ierr=0
    ext=0


    ! check input images
    nim=par%n_bkgimages



    compress_flag=par%compress_flag

    if (par%n_expimages /= nim) then
      errstr = 'Number of exposure maps and images differ'
      call WRERR(rtname,ierr,errstr)
    endif


! get image std headers:
    if (ierr < 3) then
      do i=1,nim
        call fget_stdhdr(par%bkgimages(i),image_stdhdr(i),status)   
      enddo
    endif


 !    call sort_ima(image_stdhdr,nim,status)
 !    -> moved after reading of detection masks headers

! get detection mask headers and determine image buffer size:


!    if (compress_flag) then
    nmask=0
    if (par%detmask_flag) then
!     ximbuf=0
!     yimbuf=0
      nmask=par%n_detmasks
      do i=1,par%n_detmasks
         call fget_dethdr(par%detmasks(i),mask_dethdr(i),status)
         call fget_stdhdr(par%detmasks(i),mask_stdhdr(i),status)
!         ximbuf=max(ximbuf,mask_dethdr(i)%xdatamax-mask_dethdr(i)%xdatamin+1)
!         yimbuf=max(yimbuf,mask_dethdr(i)%ydatamax-mask_dethdr(i)%ydatamin+1)
      enddo
      if (compress_flag) then
        ximbuf=0
        yimbuf=0
        do i=1,par%n_detmasks
          ximbuf=max(ximbuf,mask_dethdr(i)%xdatamax-mask_dethdr(i)%xdatamin+1)
          yimbuf=max(yimbuf,mask_dethdr(i)%ydatamax-mask_dethdr(i)%ydatamin+1)
        enddo
      endif
    else
      if (compress_flag) then
        compress_flag=.false.
        ierr = 1
        errstr = 'Need detection masks for compressed reading'
        call WRERR(rtname,ierr,errstr) 
      endif
    endif
!    endif



 !  create index arrays IP and EXPIP   
    call sort_ima(image_stdhdr,mask_stdhdr,nim,nmask,status)
!    write(*,*) 'nim,num_exp:',nim,num_exp

! read images:
    if (ierr < 3) then
      do i=1,num_exp
        ixb0(i)=0
        iyb0(i)=0
!        write(*,*) 'detmask_flag:',par%detmask_flag
        if (par%detmask_flag) then
          call fget_l1compress(par%detmasks(i),num_exp,ximbuf,yimbuf,mask_dethdr(i),compress_flag,detmask,ord,i,status)
!          write(*,*) 'read detmask #:',i 
          if (compress_flag) then
             ixb0(i)=mask_dethdr(i)%xdatamin-1
             iyb0(i)=mask_dethdr(i)%ydatamin-1
             ! set borders to .false.  to avoid problems with ixb/iyb mechanism:
             detmask%arr(i,1,:)=.false.
             detmask%arr(i,ximbuf,:)=.false.
             detmask%arr(i,:,1)=.false.
             detmask%arr(i,:,yimbuf)=.false.
          else
          ! Make sure data boundaries are present in any case:
            detmask%dethdr(i)%xdatamin = 1
            detmask%dethdr(i)%xdatamax =  detmask%shapehdr(i)%naxis1
            detmask%dethdr(i)%ydatamin = 1
            detmask%dethdr(i)%ydatamax =  detmask%shapehdr(i)%naxis2
          endif
        endif  


        do j=1,num_image(i)
          ipoint=ip(i,j)
            ixb0(i)=mask_dethdr(i)%xdatamin-1
            iyb0(i)=mask_dethdr(i)%ydatamin-1
!            call fget_i4compress(par%images(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
!                                 compress_flag,image,ord,ipoint,status)
!            if (.not. compress_flag) then
!              ximbuf =  image%shapehdr(1)%naxis1
!              yimbuf =  image%shapehdr(1)%naxis2 
!            endif
!            if (status > 0) then
!              ierr=3
!              errstr = 'Cannot read science image'
!              call WRERR(rtname,ierr,errstr) 
!            endif
            
!            if (par%expima_flag) then
              call fget_r4compress(par%expimages(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                   compress_flag,expimage,ord,ipoint,status)


!            endif
            call fget_r4compress(par%bkgimages(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                 compress_flag,bkgimage,ord,ipoint,status)
            if (status > 0) then
              ierr=3
              errstr = 'Cannot read background map'
              call WRERR(rtname,ierr,errstr)  
            else if (.not. (match_wcs(bkgimage%wcshdr(ipoint),expimage%wcshdr(ipoint))  &
                     .and. match_size(bkgimage%shapehdr(ipoint),expimage%shapehdr(ipoint)))) then
              ierr=3
              errstr = 'Mismatch of image and background map'
              call WRERR(rtname,ierr,errstr)   
            endif
!          else
!            ierr=3
!            errstr = 'Mismatch of image and detector mask'
!            call WRERR(rtname,ierr,errstr)
!          endif
        enddo
      enddo
    endif

!---------------------------------------------------------------------------------------------------- 


! read events and attitude extensions, store in evtool structure:

!   if (par%photon_flag) then
!
!      do i=1,nim
!        call open_eventfile(eventlist(i), par%images(i), READWRITE, status)
!        call read_events_tEventFile(eventlist(i), status)
!        call read_stdheader_tEventFile(eventlist(i), status)
!        call read_corrat_tEventFile(eventlist(i), status)
!        call close_eventfile(eventlist(i), status)
!        ! calcluate image pixels for events:
!        call skypix2ima(image%wcshdr(nim),int(eventlist(i)%events%x,kind=int32),int(eventlist(i)%events%y,kind=int32),pixlist(i),status)
 !     enddo
!
!    endif


  end subroutine ersensmap_in

!--------------------------------------------------------------------------------



  subroutine ersensmap_out(par,sensmap,areatab,ierr)

  !----------------------------------------------------------------------------
  ! Write ermldet output files:
  !  - output source list
  !  - source map 
  !-----------------------------------------------------------------------------

    type(ersensmap_param_type),intent(in)      :: par

    type(r4image_type)                       :: sensmap
    type(area_table_type)                    :: areatab

    integer(kind=int32)                      :: ierr
    integer(kind=int32)                      :: status,i,xsize,ysize
 
    character(len=*), parameter :: rtname='ERSENSMAP_OUT'  

    ierr = 0
    status = 0

!    write(*,*) 'NCOL=',areatab%ncol


    xsize=size(sensmap%arr,1)
    ysize=size(sensmap%arr,2)
!    allocate(sensmap%arr(xsize,ysize))
  

    status=0


!      srcmap%stdhdr=bkgimage%stdhdr(i)
!      srcmap%refhdr=bkgimage%refhdr(i)
!      srcmap%wcshdr=bkgimage%wcshdr(i)
       

!      srcmap%shapehdr%bitpix=-32
!      srcmap%shapehdr%naxis=2
!      srcmap%shapehdr%naxis1=xsize
!      srcmap%shapehdr%naxis2=ysize  

!      srcmap%stdhdr%hduclas1='IMAGE'
!      srcmap%stdhdr%hduclas2='GENERIC'
!      srcmap%stdhdr%creator='ERMLDET'

!     write(*,*) '**** write sensmap:'

      call fput_r4image(par%sensimage,sensmap,status)
      if (status >  0) then
        ierr = 3
        call wrerr(rtname,ierr,'Error writing sensitivity map file')
      endif     

!     write(*,*) '**** write area table:'


      if (par%area_flag) then
        areatab%stdhdr=sensmap%stdhdr
        call fput_areatab(par%area_table,.false.,areatab,status)
       if (status >  0) then
        ierr = 3
        call wrerr(rtname,ierr,'Error writing area table')
      endif     
        
      endif   

!    deallocate(srcmap%arr)
    

  end subroutine ersensmap_out

!---------------------------------------------------------------------------------

  subroutine ersensmap_psf(par,bkgimage,expimage,detmask,sensmap,ierr)

  !--------------------------------------------------------------------
  ! Calculate sensitivity map based on aperture counts 
  !  - get eef from aperture size / aperture size from eef
  !  - 
  !--------------------------------------------------------------------

  type(ersensmap_param_type)                 :: par
  type(l1cube_type)                          :: detmask
  type(r4cube_type)                          :: expimage
  type(r4cube_type)                          :: bkgimage
  type(r4cube_type)                          :: srcimage
  type(r4image_type)                         :: sensmap
  integer(kind=int32)                        :: ierr

  type(FPSF)                                 :: cluPSF  
  type(i4cube_type)                          :: image 

  type(ermldet_param_type)                   :: ermlpar
  real(kind=real64)                          :: sig,x0,y0
  real(kind=real32),dimension(max_ima)       :: eef,eef2,apsize,ap_area
  real(kind=real32), dimension(max_ima)      :: expo,bkg,ecf
  real(kind=real32)                          :: flux
!  real(kind=real32),dimension(:,:,:),allocatable  :: fpsf

  real(kind=real32), dimension(:,:), pointer :: fima
  integer(kind=int32)                        :: ix,iy,iim,ndof,ipoint,imask
  integer(kind=int32)                        :: i,j,isize,nstep,npix,ns,nbox
  integer(kind=int32)                        :: ix1,ix2,iy1,iy2

  integer(kind=int32)                        :: nima,iima,iinst
  real(kind=real32)                          :: sens,xycen,r2,x2,y2,diff,ecut
  logical(kind=log32)                        :: survmode
  logical(kind=log32)                        :: newcal
  logical(kind=1), dimension(:,:), allocatable  :: circmask,ormask
  real(kind=real32), dimension(:),allocatable   :: yfit1
  real(kind=real32), dimension(:,:),allocatable :: yfit,yfit0
  integer(kind=int32)                         :: status

  character(len=*), parameter :: rtname='ERSENSMAP_PSF'  

  ierr=0
  status=0
  nima=par%n_bkgimages
  ecut=15.0 ! make his a task parameter?
  isize=32
  xycen=16.5
  r2=ecut*ecut

  if (par%detmask_flag) then
    allocate(ormask(detmask%shapehdr(1)%naxis1,detmask%shapehdr(1)%naxis2))
    ormask(:,:) = .false.
    do i=1,num_exp
      do j=1,num_image(i)
        ipoint=ip(i,j)
        imask=mask_ip(ipoint)
        ormask = ormask .or. detmask%arr(:,:,imask)
      end do
    end do
  endif


  allocate(yfit1(isize*isize))
  allocate(yfit0(isize*isize,nima))
  allocate(yfit(isize*isize,nima))

  ! image is only needed to pass the header data to routine ftpsf:
  allocate(image%stdhdr(nima))
  allocate(image%wcshdr(nima))


  image%stdhdr=bkgimage%stdhdr
  image%wcshdr=bkgimage%wcshdr


  allocate(circmask(isize,isize))
  circmask(:,:)=.false.
  do j=1,isize
    diff=(float(j)-xycen)
    y2=diff*diff
    do i=1,isize
       diff=float(i)-xycen
       x2=diff*diff
       if ( (x2+y2) < r2) then
         circmask(i,j)=.true.
       endif
    enddo
  enddo


 
  



  ermlpar%shapelet_flag = par%shapelet_flag
  ermlpar%extentmodel   = par%extentmodel
  ermlpar%emin          = par%emin
  ermlpar%emax          = par%emax

  newcal=.true.
 

  survmode=.false.
  ! Use survey averaged PSF for SURVEY and SCAN modes
  if (bkgimage%stdhdr(1)%obs_mode == 'SURVEY' .or. bkgimage%stdhdr(1)%obs_mode =='SCAN' .or. bkgimage%stdhdr(1)%obs_mode =='SLEW') then
    survmode = .true.
  else if  (bkgimage%stdhdr(1)%obs_mode == 'survey' .or. bkgimage%stdhdr(1)%obs_mode =='scan' .or. bkgimage%stdhdr(1)%obs_mode =='slew') then
    survmode = .true.
  else if  (bkgimage%stdhdr(1)%obs_mode == 'POINTING'  .or. bkgimage%stdhdr(1)%obs_mode == 'pointing')  then
    ierr=1 
    call wrerr(rtname,ierr,'Warning: using survey PSF')
  endif    


  yfit0(:,:)=0.0
  ! for survey/scan mode, use global PSF:
  if (survmode) then
    iinst=1
    x0=0.D0
    y0=0.D0
    sig=0.D0
    nstep=9
    nbox=4
    ns=1
    if (par%ext_flag) sig=par%ext
    ecut=15.0 ! make this a task parameter?
    isize=32
    do iima=1,nima
      newcal=.true.
      call ftpsf(ermlpar,image,newcal,iinst,iima,ns,x0,y0,sig,ecut,cluPSF,fima,isize,status)
      yfit1 = pack(fima,circmask)
      npix = size(yfit1)
      yfit0(1:npix,iima)=yfit1
    enddo
  endif


  sensmap%arr=0.0
!  write(*,*) sensmap%shapehdr%naxis2,sensmap%shapehdr%naxis1
  do iy=1,sensmap%shapehdr%naxis2,nstep
!    write(*,*) 'row=',iy,flux
    do ix=1,sensmap%shapehdr%naxis1,nstep
        ndof=0
        iim=0
        do i=1,num_exp
          do j=1,num_image(i)
            ipoint=ip(i,j)
            imask=mask_ip(ipoint)
            if (detmask%arr(ix,iy,imask)) then
              ndof=ndof+1
              iim=iim+1
              expo(iim)=expimage%arr(ix,iy,ipoint)
              bkg(iim)=bkgimage%arr(ix,iy,ipoint)
              eef2(iim)=eef(ipoint)
              ecf(iim)=par%ecf(ipoint)
              yfit(:,iim)=yfit0(:,ipoint)
            endif
          enddo 
          if (ndof > 0) then
            call iterate_cash(par,bkg,expo,ecf,ndof,yfit0,npix,flux,status)
            ix1=max(1,ix-nbox)
            ix2=min(sensmap%shapehdr%naxis1,ix+nbox)
            iy1=max(1,iy-nbox)
            iy2=min(sensmap%shapehdr%naxis2,iy+nbox)

!            sensmap%arr(ix,iy)=flux
            sensmap%arr(ix1:ix2,iy1:iy2)=flux
          endif
      enddo
    enddo
  enddo

  do iy=1,sensmap%shapehdr%naxis2
    do ix=1,sensmap%shapehdr%naxis1
      if (.not. ormask(ix,iy)) sensmap%arr(ix,iy)=0.0
    end do
  end do

  end subroutine ersensmap_psf

!---------------------------------------------------------------------------------


  subroutine  iterate_cash(par,bkg,expo,ecf,ndof,yfit,npix,flux,ierr)

    !-----------------------------------------------------------------
    ! Iteratively calculate fluxes needed to reach the limiting 
    ! likelihood  par%likemin using the C-statisyic.
    ! Inputs are background levels (bkg), exposure (expo), 
    ! energy conversion factors (ecf), and the spatial 
    ! source distribution (yfit) for each of the <nima> input images.  
    ! -----------------------------------------------------------------

    type(ersensmap_param_type), intent(in)          :: par
    real(kind=real32), dimension(:)                 :: expo
    real(kind=real32), dimension(:)                 :: bkg
    real(kind=real32), dimension(:)                 :: ecf
    integer(kind=int32)                             :: ndof
    real(kind=real32), dimension(:,:),allocatable   :: yfit
    integer(kind=int32)                             :: npix
    real(kind=real32)                               :: flux
    integer(kind=int32)                             :: ierr
    real(kind=real32), dimension(:,:),allocatable   :: yfit_scal

    integer(kind=int32)                             :: status
    real(kind=real32)                               :: counts_old,counts,cstep,eps
    real(kind=real32)                               :: dc,ldiff,like,like_old
    real(kind=real32)                               :: flux_old,df,nu2
    integer(kind=int32)                             :: nim,i,iim,ipix
    real(kind=real32)                               :: btot,etot

    character(len=*), parameter                     :: rtname='ITERATE_CASH'
    character(len=80)                               :: errstr
    
    
    ierr=0
    status=0
    eps=1.E-3
    nim=ndof
   ! write(*,*) 'allocate yfit_scal:',npix,nim
    allocate(yfit_scal(npix,nim))

    if (par%ext_flag) then 
      nu2=real(ndof+3)/2.
    else
      nu2=real(ndof+2)/2.
    endif      

    counts_old=5.0
    flux_old=0.0
    like_old=0.0

    do iim=1,nim
      flux_old = flux_old + counts_old/expo(iim)/real(nim)/ecf(iim)
      yfit_scal(:,iim) = yfit(1:npix,iim)*flux_old*ecf(iim)*expo(iim)+bkg(iim)
      btot=bkg(iim)*npix
      etot=sum(yfit_scal(:,iim))
      like_old = 0.0 
      do ipix=1,npix
      !  write(*,*) 'alog args:',yfit_scal(ipix,iim),bkg(iim)
        like_old = like_old+yfit_scal(ipix,iim)*(alog(yfit_scal(ipix,iim))-alog(bkg(iim)))
      end do
      like_old=2.*(like_old+btot-etot)
      !like_old = like_old + gammp(flux_old*ecf(iim)*eef(iim)*expo(iim),bkg(iim)) 
    enddo

    
    like_old = gammq(nu2, like_old) 
    df=0.2*flux_old
    

    do i=1,100
      like=0.0
      flux=flux_old+df
      do iim=1,nim 
        yfit_scal(:,iim) = yfit(1:npix,iim)*flux*ecf(iim)*expo(iim)+bkg(iim)
        etot=sum(yfit_scal(:,iim))
        ! 07/08/2019 GL:
        ! like = 0.0 
        do ipix=1,npix
          like = like+yfit_scal(ipix,iim)*(alog(yfit_scal(ipix,iim))-alog(bkg(iim)))
        end do
        like = 2.*(like+btot-etot)        
      enddo
      like = gammq(nu2, like) 
      ldiff = par%likemin-like
      if (abs(ldiff) <  eps) exit
     ! write(*,*) 'like:',like,like_old
      df  = df*ldiff/(like-like_old) 
      like_old = like
      flux_old = flux
    enddo
    
    deallocate(yfit_scal)

    if (i > 99) then
      ierr=3
      errstr='iteration failed'
      call wrerr(rtname,ierr,errstr)
    endif
     


  end subroutine  iterate_cash 



!---------------------------------------------------------------------------------


  subroutine ersensmap_calc(par,bkgimage,expimage,detmask,sensmap,ierr)

  !--------------------------------------------------------------------
  ! Calculate sensitivity map based on aperture counts 
  !  - get eef from aperture size / aperture size from eef
  !  - 
  !--------------------------------------------------------------------

  type(ersensmap_param_type)                 :: par
  type(l1cube_type)                          :: detmask
  type(r4cube_type)                          :: expimage
  type(r4cube_type)                          :: bkgimage
  type(r4cube_type)                          :: srcimage
  type(r4image_type)                         :: sensmap
  integer(kind=int32)                        :: ierr

  real(kind=real32),dimension(max_ima)       :: eef,eef2,apsize,ap_area
  real(kind=real32), dimension(max_ima)      :: expo,bkg,ecf
  real(kind=real32)                          :: flux
  type(psf2d_type),dimension(:),allocatable  :: psf
  integer(kind=int32)                        :: ix,iy,iim,ndof,ipoint,imask
  integer(kind=int32)                        :: i,j
  integer(kind=int32)                        :: status,nima,iima
  real(kind=real32)                          :: sens
  logical(kind=log32)                        :: survmode

  ierr=0
  status=0
  nima=par%n_bkgimages
  
  allocate(psf(nima))


!  survmode=.false.
  survmode=.true.
  ! Use survey averaged PSF for SURVEY and SCAN modes
  if (bkgimage%stdhdr(1)%obs_mode == 'SURVEY' .or. bkgimage%stdhdr(1)%obs_mode =='SCAN' .or. bkgimage%stdhdr(1)%obs_mode =='SLEW' ) then
    survmode = .true.
  else if (bkgimage%stdhdr(1)%obs_mode == 'survey' .or. bkgimage%stdhdr(1)%obs_mode =='scan' .or. bkgimage%stdhdr(1)%obs_mode =='slew' ) then
    survmode = .true.
  endif    

  ! for survey/scan mode, use global PSF:
  if (survmode) then
    call make_aper(par,bkgimage,survmode,eef,apsize,ap_area,psf,status)
  endif
! write(*,*) 'apsize:',apsize(1:2),eef(1:2),ap_area(1:2),ecf(1:2)

  sensmap%arr=0.0
!  write(*,*) sensmap%shapehdr%naxis2,sensmap%shapehdr%naxis1
  do iy=1,sensmap%shapehdr%naxis2
    do ix=1,sensmap%shapehdr%naxis1
        ndof=0
        iim=0
        do i=1,num_exp
          do j=1,num_image(i)
            ipoint=ip(i,j)
            imask=mask_ip(ipoint)
            if (detmask%arr(ix,iy,imask)) then
              ndof=ndof+1
              iim=iim+1
              expo(iim)=expimage%arr(ix,iy,ipoint)
              bkg(iim)=bkgimage%arr(ix,iy,ipoint)*ap_area(ipoint)
              eef2(iim)=eef(ipoint)
              ecf(iim)=par%ecf(ipoint)
            endif
          enddo 
!          write(*,*) ix,iy
          if (ndof > 0) then
!            write(*,*) "start mgamma:",bkg(1)
            call iterate_mgamma(par,bkg,expo,ecf,ndof,eef2,flux,status)
            sensmap%arr(ix,iy)=flux
          endif
      enddo
    enddo
  enddo

  end subroutine ersensmap_calc

!---------------------------------------------------------------------------------


  subroutine make_aper(par,image,survmode,eef,apsize,ap_area,psf,ierr)

  !--------------------------------------------------------------
  !  Determine source cut-out radii from PSF 
  ! 
  !  par%aper_size : aperture size (radius|box_hw) for sub-image used for fitting 
  !  rad < 1.0  : interpret as PSF encircled energy fraction
  !               -> calculate corresponding radius in image pixels
  !  rad >= 1.0 : interpret as radius in pixels
  !               -> calculate EEF 
  !  Return eef, apsize, psf 
  !-------------------------------------------------------------- 


    type(ersensmap_param_type)                 :: par
    type(r4cube_type)                          :: image
    logical (kind=log32)                       :: survmode
    real(kind=real32),dimension(:)             :: eef
    real(kind=real32),dimension(:)             :: apsize
    real(kind=real32),dimension(:)             :: ap_area

    integer(kind=int32)  :: ierr
    type(cal_par_type)   :: cal_par
    type(psf2d_type),dimension(:),allocatable     :: psf
    real(kind=real32)    :: pi=3.141592654, twopi=6.283185308
    real(kind=real32)    :: energy
    integer(kind=int32)  :: ext,ext_old    
    character(len=clong) :: psf_file,psf_file_old
    integer(kind=int32)  :: i,j,k,ipoint,status,nima,iima
    logical(kind=log32)  :: boxflag
    real(kind=real32)    ::  radius,offax,phi,deltax,deltay,pix2arcsec
    real(kind=real64)    ::  xcen,ycen

    real(kind=real64)    ::  E, theta, fraction

    character(len=*),parameter   :: rtname = 'make_aper'


    ierr=0
    nima=par%n_bkgimages
    boxflag=.false.
    if (par%aper_type == 'BOX' .or. (par%aper_type == 'box'))  boxflag=.true.

 
    psf_file_old='NONE'
    ext_old=-99
    do iima=1,nima
      energy=0.5*(par%emin(iima)+par%emax(iima))
      pix2arcsec = 0.5*(abs(image%wcshdr(1)%cdelt1)+abs(image%wcshdr(1)%cdelt2))*3600.
      ! Note: in case of NINST>1, cal_setpar uses INSTRUM1 
      call cal_setpar(image%stdhdr(iima),cal_par,status) 
      if (survmode) then
        phi=0.0
        ! get survey PSF
        cal_par%codenam='2D_PSF'
        cal_par%expression=''
        
        call cal_r4toex('ENERG',energy,cal_par%expression,status)
        call cal_stringtoex('OBSMODE','SURVEY',cal_par%expression,status)


        call cal_filename(cal_par,psf_file,ext,status)
        call get_2dpsf(psf_file,ext,phi,psf(iima),status)
        ! use same radii for all sources:

        if (par%aper_size < 1.0) then
          if (boxflag) then
            call psf_eefbox(psf(iima),par%aper_size,apsize(iima),status)
          else
            call  psf_eefrad(psf(iima),par%aper_size,apsize(iima),ierr)
          endif
        else
          apsize(iima)=par%aper_size
          if (boxflag) then
!            write(*,*) "psf:",iima,par%aper_size*pix2arcsec
            call psf_boxfrac(psf(iima),par%aper_size*pix2arcsec,eef(iima),status) 

          else
            call psf_eef(psf(iima),par%aper_size*pix2arcsec,.false.,eef(iima),status)
          endif
        endif   
        if (boxflag) then
          ap_area(iima)=4.*apsize(iima)*apsize(iima)
        else
          ap_area(iima)=pi*apsize(iima)*apsize(iima)
        endif
      else ! end survmode - pointed mode:


    !    call  equ2pix(image%wcshdr(iima),image%stdhdr(iima)%ra_pnt,image%stdhdr(iima)%dec_pnt,xcen,ycen,status)
    !    do i=1,boxlist%shapehdr%naxis2
    !      deltax=abs(boxlist%data(0,i)%x_ima-xcen)
    !      deltay=abs(boxlist%data(0,i)%y_ima-ycen)
    !      offax = sqrt((deltax*image%wcshdr(1)%cdelt1)**2+(deltay*image%wcshdr(1)%cdelt2)**2)*60.
    !      offax = sqrt((deltax*image%wcshdr(1)%cdelt1)**2+(deltay*image%wcshdr(1)%cdelt2)**2)*60.
    
    !      if (deltax == 0.0 .and. deltay == 0.0) then
    !        phi = 0.0
    !      else
    !        phi = atan2(deltax,deltay)
    !      endif
    !      cal_par%expression=''
   !       call cal_r4toex('ENERG',energy,cal_par%expression,status)
   !       call cal_r4toex('THETA',offax,cal_par%expression,status)
   !       call cal_stringtoex('OBSMODE','POINTING',cal_par%expression,status)
   !       call cal_filename(cal_par,psf_file,ext,status)
    !      if (psf_file /= psf_file_old .or. ext /= ext_old) then
    !        call get_2dpsf(psf_file,ext,phi,psf,status)   
    !      endif
    !      psf_file_old=psf_file
    !      ext_old=ext   
    !         
    !      if (par%multrad < 1.0) then
    !        call  psf_eefrad(psf,par%multrad,radius,ierr)
    !        if (radius > boxlist%data(iima,i)%ecutrad) then
    !          boxlist%data(:,i)%ecutrad = radius
    !        endif
    !      else
    !         boxlist%data(:,i)%ecutrad = par%aper_size
    !!         call psf_eef(psf,par%aper_size,.false.,eef,status)
    !         boxlist%data(iima,i)%eef=eef
    !      endif
    !      if (par%aper_size < 1.0) then
    !        call  psf_eefrad(psf,par%multrad,radius,ierr)
    !        if (radius > boxlist%data(iima,i)%ecutrad) then
    !          boxlist%data(:,i)%ecutrad = radius
    !        endif
    !      else
    !        boxlist%data(iima,i)%ecutrad = par%aper_size
    !        call psf_eef(psf,par%aper_size,.false.,eef,status)
    !        boxlist%data(iima,i)%eef=eef
    !      endif
    !    enddo

      endif ! end pointed mode


    end do

    ! 02/06/2017: Don't deallocate here, psf is returned to calling routine
    ! if (associated(psf%arr)) deallocate(psf%arr)


  end subroutine make_aper

!-------------------------------------------------------------------------------

  subroutine  iterate_gamma(par,bkg,counts,ierr)

    type(ersensmap_param_type), intent(in)          :: par
    real(kind=real32)                               :: bkg
    real(kind=real32)                               :: counts
    integer(kind=int32)                             :: ierr
    integer(kind=int32)                             :: status
    real(kind=real32)                               :: counts_old,cstep,eps
    real(kind=real32)                               :: dc,ldiff,like,like_old
    integer(kind=int32)                             :: i,iim
    character(len=*), parameter                     :: rtname='ITERATE_GAMMA'
    character(len=80)                               :: errstr
    
    
    ierr=0
    status=0
    eps=1.E-3
   
    counts_old=5.0
    dc=1.0
    like_old = gammp(counts_old,bkg) 

    do i=1,100
      write(*,*) "iterate:",i,counts_old,dc
      counts = counts_old+dc
      like = gammp(counts,bkg) 
      ldiff = par%likemin-like 

      if (abs(ldiff) <  eps) exit
      dc  = dc*ldiff/(like-like_old) 
      like_old = like
      counts_old = counts
    enddo
    
    if (i > 99) then
      ierr=3
      errstr='iteration failed'
      call wrerr(rtname,ierr,errstr)
    endif
    
  end subroutine iterate_gamma


  !---------------------------------------------------------------------


  !---------------------------------------------------------------------

  subroutine  iterate_mgamma(par,bkg,expo,ecf,ndof,eef,flux,ierr)

  !------------------------------------------------------------------
  ! as iterate_gamma, but accepts multiple sets of bkg,exp values
  ! and ecf values, output is minimum source flux
  !------------------------------------------------------------------

    type(ersensmap_param_type), intent(in)          :: par
    real(kind=real32), dimension(:)                 :: expo
    real(kind=real32), dimension(:)                 :: bkg
    real(kind=real32), dimension(:)                 :: ecf
    real(kind=real32), dimension(:)                 :: eef
    real(kind=real32)                               :: flux
    integer(kind=int32)                             :: ierr
    integer(kind=int32)                             :: status
    real(kind=real32)                               :: counts_old,counts,cstep,eps
    real(kind=real32)                               :: dc,ldiff
    real(kind=real32)                               :: df,nu2,like
    real(kind=real32), save                         :: like_old, flux_old
    integer(kind=int32)                             :: nim,i,iim,ndof

    logical(kind=log32) , save                      :: first=.true.
    character(len=*), parameter                     :: rtname='ITERATE_MGAMMA'
    character(len=80)                               :: errstr
    
    
     if (first) then
       
       counts_old=20.0
       flux_old=0.0
       like_old=0.0 

     endif


    ierr=0
    status=0
    eps=1.E-3
    nim=ndof
    ! check whether this is correct: ! 
    nu2=real(ndof)/2.



    if (first) then
      counts_old=20.0
      flux_old=0.0
      like_old=0.0
      do iim=1,nim
        flux_old = flux_old + counts_old/expo(iim)/real(nim)/ecf(iim)/eef(iim)
!       write(*,*) 'flux:',flux_old,ecf(iim),eef(iim),expo(iim),bkg(iim)
!       write(*,*) "gammp 1st:",gammp(flux_old*ecf(iim)*eef(iim)*expo(iim),bkg(iim))
        like_old = like_old + gammp(flux_old*ecf(iim)*eef(iim)*expo(iim),bkg(iim))
    
        like_old = gammq(nu2, like_old) 
          
       enddo
       first=.false.
    endif
    df=0.2*flux_old

    do i=1,100
      like=0
      flux=flux_old+df
      do iim=1,nim 
!        write(*,*) 'mgamma 1 :',flux,ecf(iim),eef(iim),expo(iim),bkg(iim)
!        write(*,*) 'mgamma 2:', flux*ecf(iim)*eef(iim)*expo(iim), bkg(iim)  
!       write(*,*) "gammp loop:",gammp(flux*ecf(iim)*eef(iim)*expo(iim),bkg(iim))
        like = like+gammp(flux*ecf(iim)*eef(iim)*expo(iim),bkg(iim))
         
      enddo
      like = gammq(nu2, like) 
      ldiff = par%likemin-like
      if (abs(ldiff) <  eps) exit
      if ((like-like_old) .ne. 0.0) then
        df  = df*ldiff/(like-like_old) 
      else
        df=sign(eps,ldiff)
      endif
      like_old = like
      flux_old = flux
    enddo
    
!    stop

    if (i > 99) then
      ierr=2
      errstr='iteration failed'
      call wrerr(rtname,ierr,errstr)
    endif
    
  end subroutine iterate_mgamma


  !---------------------------------------------------------------------



  !---------------------------------------------------------------------

  subroutine  cash(  )


  end subroutine cash


  !---------------------------------------------------------------------

  !---------------------------------------------------------------------

  subroutine  area_calc(par,sensmap,detmask,areatab,ierr)

  type(ersensmap_param_type)                 :: par
  type(r4image_type)                         :: sensmap
  type(l1cube_type)                          :: detmask
  type(area_table_type)                      :: areatab
  integer(kind=int32)                        :: ierr

  integer(kind=int32)                        :: status
  integer(kind=int32)                        :: nbins=101,igrid
   integer(kind=int32)                       :: ix,iy 
  
  real(kind=real32)                          :: pixarea
  real(kind=real64)                          :: gridfact
  real(kind=real64)                          :: logmin,logfact=0.05D0
  real(kind=real32)                            :: minflux=1.E-16,maxflux=1E-11
  integer(kind=int32),dimension(:),allocatable :: histogram
  real(kind=real32),dimension(:),allocatable   :: flimits
  !test:
  integer(kind=int32)                          :: totcount,selcount,imcount 
 
  
  character(len=*), parameter                 :: rtname='AREA_CALC'




  ierr=0
  status=0
  imcount=0
  totcount=0
  selcount=0
    

  pixarea=abs(sensmap%wcshdr%cdelt1*sensmap%wcshdr%cdelt2)
  logmin=alog10(minflux)
  gridfact=10.D0**logfact
  allocate(flimits(nbins))
  allocate(histogram(nbins))


  do igrid=1,nbins
    flimits(igrid)=minflux*gridfact**(dble(igrid))
  enddo

  histogram(:) = 0   
  do iy=1,sensmap%shapehdr%naxis2
    do ix=1,sensmap%shapehdr%naxis1
      imcount=imcount+1
      if (sensmap%arr(ix,iy) > 0.0) then
        igrid=floor( (alog10(sensmap%arr(ix,iy))-logmin)/logfact )
        totcount=totcount+1
        if (igrid > 0 .and. igrid <= nbins) then
           selcount=selcount+1
           histogram(igrid) = histogram(igrid) + 1
        endif
       endif
    enddo
  enddo 

  allocate(areatab%data(nbins))
  

  areatab%data(1)%limflux=flimits(1)
  areatab%data(1)%sky_area=histogram(1)*pixarea

  do igrid=2,nbins
   areatab%data(igrid)%limflux=flimits(igrid)
   areatab%data(igrid)%sky_area=areatab%data(igrid-1)%sky_area+float(histogram(igrid))*pixarea
  enddo
  areatab%shapehdr%naxis2=nbins 


  end subroutine area_calc


  !---------------------------------------------------------------------




!--------------------------------------------------------------------------------

  subroutine fput_areatab(filename,append,area_tab,ierr)

    character(len=*), intent(in)       :: filename
    logical(kind=log32)                :: append   
    type(area_table_type)              :: area_tab
    integer(kind=int32), intent(out)   :: ierr
    logical                            :: fnull,simple
    integer(kind=int32)                :: nrows,maxbuf,irow,rowbuf,i1,i2
    integer(kind=int32)                :: status,fstat,unit,blocksize,hdutype,rwmode
    integer(kind=int32)                :: nhdu, varidat
    integer(kind=int32)                :: bitpix,naxis,naxes(1)   
    integer(kind=int32), parameter     :: ncol=2
    character(len=16), dimension(ncol) :: ttype,tunit
    character(len=4), dimension(ncol)  :: tform
    character(len=*),parameter         :: rtname = 'FPUT_AREATAB'
    character(len=80)                  :: errstr



    status = 0
    fstat = 0
    ierr = 0
    rwmode = 1  ! readwrite
    hdutype = 2    ! binary table
    fnull  = .false.
    blocksize = 1
    varidat=0
    nrows=area_tab%shapehdr%naxis2
    
 
    call FTGIOU(unit,status)
    fstat=max(fstat,status)
    if (append) then
      call FTOPEN(unit,filename,rwmode,blocksize,status)
      fstat=max(fstat,status)
    else
      call FTINIT(unit,filename,blocksize,status)
      fstat=max(fstat,status)
      if (fstat == 0) then
        status=0
        simple   = .true.
        bitpix   = 16
        naxis    = 0
        naxes(1) = 0
        !write FITS standard header:   
        call FTPHPS(unit,bitpix,naxis,naxes,status) 
        fstat=max(fstat,status)  
        ! write eROSITA standard header to primary extension:
        call fwrite_stdhdr(unit,area_tab%stdhdr,status)
      endif
    endif

    if (fstat > 0) then
      ierr = 3
      errstr = 'error opening file '//trim(filename)
      call WRERR(rtname,ierr,errstr)
    else 
      status=0
      call FTCRHD(unit,status)
      status=0
!      write(*,*) nrows,area_tab%ncol,area_tab%cname,area_tab%cform
      call FTPHBN(unit,nrows,area_tab%ncol,area_tab%cname,area_tab%cform,area_tab%cunit,area_tab%extname,varidat,status)
      fstat=max(fstat,status)  
      

      call fwrite_stdhdr(unit,area_tab%stdhdr,status)
      fstat=max(fstat,status)


      if (fstat /= 0) then   
        ierr = 3
        errstr = 'error writing file '//trim(filename)
        call WRERR(rtname,ierr,errstr) 
      else 
        ! get maximum buffer size:
        call FTGRSZ(unit, maxbuf, status)
        irow=0 
        do while (irow < nrows .and. fstat == 0)
          rowbuf = min(maxbuf,nrows-irow)
          i1=irow+1
          i2=irow+rowbuf
          irow=i2  
          call  FTPCLE(unit,1,i1,1,rowbuf,area_tab%data(i1:i2)%limflux,status)
          fstat = max(fstat,status)
          call  FTPCLE(unit,2,i1,1,rowbuf,area_tab%data(i1:i2)%sky_area,status)
          fstat = max(fstat,status)

         enddo
      endif  

      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error writing values to table '//trim(filename)
        call WRERR(rtname,ierr,errstr)
      endif
    
      status=0   
      ! close and free unit
      call FTCLOS(unit,status)
      call FTFIOU(unit,status)
      if (fstat /= 0) then
        ierr = 3
        errstr = 'Error closing file '//filename
        call WRERR(rtname,ierr,errstr)
      endif
    endif  

  end   subroutine fput_areatab


  


  !---------------------------------------------------------------------

  



end module  ersensmap_mod
